# coding utf-8
import pytest

from ladybug_comfort.collection.solarcal import OutdoorSolarCal, IndoorSolarCal, \
    HorizontalSolarCal
from ladybug_comfort.parameter.solarcal import SolarCalParameter

from ladybug_comfort.solarcal import outdoor_sky_heat_exch, indoor_sky_heat_exch, \
    shortwave_from_horiz_solar, mrt_delta_from_erf, erf_from_mrt_delta, \
    get_projection_factor, get_projection_factor_simple, \
    sharp_from_solar_and_body_azimuth, body_solar_flux_from_parts, \
    body_solar_flux_from_horiz_parts

from ladybug.location import Location
from ladybug.analysisperiod import AnalysisPeriod
from ladybug.header import Header
from ladybug.datacollection import HourlyContinuousCollection
from ladybug.epw import EPW
from ladybug.wea import Wea
from ladybug.sunpath import Sunpath

from ladybug.datatype.energyflux import Irradiance

import math
import sys
if (sys.version_info > (3, 0)):
    xrange = range


def test_outdoor_sky_heat_exch():
    """Test the outdoor_sky_heat_exch function"""
    # Test typical daytime condition
    sky_exch = outdoor_sky_heat_exch(22, 380, 200, 380, 45)
    assert sky_exch['s_erf'] == pytest.approx(129.239, rel=1e-2)
    assert sky_exch['s_dmrt'] == pytest.approx(29.6508, rel=1e-2)
    assert sky_exch['l_erf'] == pytest.approx(-11.6208, rel=1e-2)
    assert sky_exch['l_dmrt'] == pytest.approx(-2.6661, rel=1e-2)
    assert sky_exch['mrt'] == pytest.approx(48.9847, rel=1e-2)

    # Test typical nighttime condition
    sky_exch = outdoor_sky_heat_exch(18, 330, 0, 0, 0)
    assert sky_exch['s_erf'] == pytest.approx(0, rel=1e-2)
    assert sky_exch['s_dmrt'] == pytest.approx(0, rel=1e-2)
    assert sky_exch['l_erf'] == pytest.approx(-24.792, rel=1e-2)
    assert sky_exch['l_dmrt'] == pytest.approx(-5.688, rel=1e-2)
    assert sky_exch['mrt'] == pytest.approx(12.3120, rel=1e-2)


def test_indoor_sky_heat_exch():
    """Test the indoor_sky_heat_exch function"""
    # Test typical daytime condition
    sky_exch = indoor_sky_heat_exch(22, 200, 380, 45, 0.5, 0.5)
    assert sky_exch['erf'] == pytest.approx(23.71407, rel=1e-2)
    assert sky_exch['dmrt'] == pytest.approx(5.66732, rel=1e-2)
    assert sky_exch['mrt'] == pytest.approx(27.6673, rel=1e-2)

    # Test typical nighttime condition
    sky_exch = indoor_sky_heat_exch(22, 0, 0, 0, 0.5, 0.5)
    assert sky_exch['erf'] == pytest.approx(0, rel=1e-2)
    assert sky_exch['dmrt'] == pytest.approx(0, rel=1e-2)
    assert sky_exch['mrt'] == pytest.approx(22, rel=1e-2)


def test_shortwave_from_horiz_solar():
    """Test the shortwave_from_horiz_solar function."""
    # Test typical daytime noon condition
    sky_exch = shortwave_from_horiz_solar(22, 144, 850, 72)
    assert sky_exch['erf'] == pytest.approx(168.7179, rel=1e-2)
    assert sky_exch['dmrt'] == pytest.approx(38.7083, rel=1e-2)
    assert sky_exch['mrt'] == pytest.approx(60.7083, rel=1e-2)

    # Test typical daytime condition
    sky_exch = shortwave_from_horiz_solar(22, 120, 500, 45)
    assert sky_exch['erf'] == pytest.approx(157.33914, rel=1e-2)
    assert sky_exch['dmrt'] == pytest.approx(36.09772, rel=1e-2)
    assert sky_exch['mrt'] == pytest.approx(58.09772, rel=1e-2)

    # Test typical daytime low angle condition
    sky_exch = shortwave_from_horiz_solar(22, 10, 55, 15)
    assert sky_exch['erf'] == pytest.approx(41.61606, rel=1e-2)
    assert sky_exch['dmrt'] == pytest.approx(9.547814, rel=1e-2)
    assert sky_exch['mrt'] == pytest.approx(31.547814, rel=1e-2)

    # Test typical nighttime condition
    sky_exch = shortwave_from_horiz_solar(18, 0, 0, -10, 0.5)
    assert sky_exch['erf'] == pytest.approx(0, rel=1e-2)
    assert sky_exch['dmrt'] == pytest.approx(0, rel=1e-2)
    assert sky_exch['mrt'] == pytest.approx(18, rel=1e-2)


def test_solarcal_validation():
    """Test the SolarCal function against the reference table from ASHRAE-55 2017.
    """
    validation_csv_file_path = './tests/validation_tables/solarcal_validation.csv'
    with open(validation_csv_file_path) as csv_data_file:
        csv_data_file.readline()
        csv_data_file.readline()
        for row in csv_data_file:
            vals = []
            for val in row.split(','):
                try:
                    vals.append(float(val))
                except ValueError:
                    vals.append(str(val))
            i_diff = 0.17 * vals[3] * math.sin(math.radians(vals[0]))
            result = indoor_sky_heat_exch(20, i_diff, vals[3], vals[0], vals[5],
                                          vals[6], 0.6, vals[4], vals[2],
                                          vals[1], vals[7])
            # Values don't match perfectly because we use the original SolarCal splines
            assert result['erf'] == pytest.approx(vals[8], rel=1e-2)
            assert result['dmrt'] == pytest.approx(vals[9], rel=1e-1)


def test_body_dir_from_dir_normal():
    """Test body_solar_flux_from_parts gainst its horizontal counterpart."""
    wea_obj = Wea.from_epw_file('./tests/epw/chicago.epw')
    diff_hr = wea_obj.diffuse_horizontal_irradiance.values
    dir_nr = wea_obj.direct_normal_irradiance.values
    dir_hr = wea_obj.direct_horizontal_irradiance.values
    dts = wea_obj.datetimes
    sp = Sunpath.from_location(wea_obj.location)

    for i in xrange(8760):
        sun = sp.calculate_sun_from_date_time(dts[i])
        alt, az = sun.altitude, sun.azimuth
        sharp = sharp_from_solar_and_body_azimuth(az, 180)
        sflux1 = body_solar_flux_from_parts(diff_hr[i], dir_nr[i], alt, sharp)
        sflux2 = body_solar_flux_from_horiz_parts(diff_hr[i], dir_hr[i], alt, sharp)
        assert sflux1 == pytest.approx(sflux2, rel=1e-2)


def test_mrt_delta_from_erf():
    """Test the mrt_delta_from_erf function."""
    dmrt1 = mrt_delta_from_erf(100)
    dmrt2 = mrt_delta_from_erf(0)
    dmrt3 = mrt_delta_from_erf(-100)

    assert dmrt1 == abs(dmrt3) == pytest.approx(22.94262, rel=1e-2)
    assert dmrt2 == 0


def test_erf_from_mrt_delta():
    """Test the erf_from_mrt_delta function."""
    erf1 = erf_from_mrt_delta(22.94262050611421)
    erf2 = erf_from_mrt_delta(0)
    erf3 = erf_from_mrt_delta(-22.94262050611421)

    assert erf1 == abs(erf3) == pytest.approx(100, rel=1e-2)
    assert erf2 == 0


def test_sharp_from_solar_and_body_azimuth():
    """Test the sharp_from_solar_and_body_azimuth function."""
    assert sharp_from_solar_and_body_azimuth(0, 0) == 0
    assert sharp_from_solar_and_body_azimuth(180, 180) == 0
    assert sharp_from_solar_and_body_azimuth(0, 90) == 90
    assert sharp_from_solar_and_body_azimuth(360, 90) == 90
    assert sharp_from_solar_and_body_azimuth(270, 90) == 180
    assert sharp_from_solar_and_body_azimuth(360, 180) == 180
    assert sharp_from_solar_and_body_azimuth(360, 270) == 90
    assert sharp_from_solar_and_body_azimuth(0, 270) == 90
    assert sharp_from_solar_and_body_azimuth(90, 270) == 180


def test_projection_factors():
    """Test the projection factor functions against one another."""
    for posture in ('standing', 'seated', 'supine'):
        for alt in list(xrange(1, 90, 10)) + [90]:
            for sharp in xrange(0, 190, 10):
                pf1 = get_projection_factor(alt, sharp, posture)
                pf2 = get_projection_factor_simple(alt, sharp, posture)
                assert pf1 == pytest.approx(pf2, rel=0.1)


def test_projection_factors_incorrect():
    """Test the projection factor functions with incorrect inputs."""
    with pytest.raises(Exception):
        get_projection_factor(25, 0, 'Standing')  # incorrect capitalization
    with pytest.raises(Exception):
        get_projection_factor(100, 0, 'standing')  # incorrect altitude
    with pytest.raises(Exception):
        get_projection_factor_simple(25, 0, 'Standing')  # incorrect capitalization
    with pytest.raises(Exception):
        get_projection_factor_simple(100, 0, 'standing')  # incorrect altitude


def test_solarcal_parameter_init():
    """Test the initialization of the SolarCalParameter object."""
    posture = 'seated'
    sharp = 180
    absorptivity = 0.8
    emissivity = 0.97
    solarcal_par = SolarCalParameter(posture=posture,
                                     sharp=sharp,
                                     body_absorptivity=absorptivity,
                                     body_emissivity=emissivity)
    assert solarcal_par.posture == posture
    assert solarcal_par.sharp == sharp
    assert solarcal_par.body_azimuth is None
    assert solarcal_par.body_absorptivity == absorptivity
    assert solarcal_par.body_emissivity == emissivity


def test_solarcal_parameter_default():
    """Test the default SolarCalParameter properties."""
    solarcal_par = SolarCalParameter()
    str(solarcal_par)  # test casting the parameters to a string
    assert solarcal_par.posture == 'standing'
    assert solarcal_par.sharp == 135
    assert solarcal_par.body_azimuth is None
    assert solarcal_par.body_absorptivity == 0.7
    assert solarcal_par.body_emissivity == 0.95
    solarcal_par.acceptable_postures  # test that the acceptable postures are there


def test_solarcal_parameter_incorrect():
    """Test incorrect SolarCalParameter properties."""
    with pytest.raises(Exception):
        SolarCalParameter(posture='seated',
                          sharp=180,
                          body_azimuth=135,  # both sharp and azimuth
                          body_absorptivity=0.8,
                          body_emissivity=0.97)
    with pytest.raises(Exception):
        SolarCalParameter(posture='disco pose')  # incorrect posture
    with pytest.raises(Exception):
        SolarCalParameter(sharp=270)  # incorrect sharp
    with pytest.raises(Exception):
        SolarCalParameter(body_absorptivity=60)  # incorrect body_absorptivity
    with pytest.raises(Exception):
        SolarCalParameter(body_emissivity=97)  # incorrect body_emissivity


def test_init_outdoor_solarcal_collection():
    """Test the initialization of the OutdoorSolarCal collection."""
    calc_length = 24
    irr_header = Header(Irradiance(), 'W/m2', AnalysisPeriod(end_month=1, end_day=1))
    dir_norm = HourlyContinuousCollection(irr_header, [500] * calc_length)
    diff_horiz = HourlyContinuousCollection(irr_header, [200] * calc_length)
    solarcal_obj = OutdoorSolarCal(Location(), dir_norm, diff_horiz, 350, 24)

    assert solarcal_obj.comfort_model == 'Outdoor SolarCal'
    assert solarcal_obj.calc_length == calc_length
    str(solarcal_obj)  # test that the string representaiton is ok

    assert isinstance(solarcal_obj.direct_normal_solar, HourlyContinuousCollection)
    assert len(solarcal_obj.direct_normal_solar.values) == calc_length
    assert solarcal_obj.direct_normal_solar[12] == 500
    assert isinstance(solarcal_obj.diffuse_horizontal_solar, HourlyContinuousCollection)
    assert len(solarcal_obj.diffuse_horizontal_solar.values) == calc_length
    assert solarcal_obj.diffuse_horizontal_solar[12] == 200

    assert isinstance(solarcal_obj.shortwave_effective_radiant_field, HourlyContinuousCollection)
    assert len(solarcal_obj.shortwave_effective_radiant_field.values) == calc_length
    assert solarcal_obj.shortwave_effective_radiant_field[12] == pytest.approx(137.30710, rel=1e-3)
    assert isinstance(solarcal_obj.longwave_effective_radiant_field, HourlyContinuousCollection)
    assert len(solarcal_obj.longwave_effective_radiant_field.values) == calc_length
    assert solarcal_obj.longwave_effective_radiant_field[12] == pytest.approx(-28.8326, rel=1e-3)
    assert isinstance(solarcal_obj.shortwave_mrt_delta, HourlyContinuousCollection)
    assert len(solarcal_obj.shortwave_mrt_delta.values) == calc_length
    assert solarcal_obj.shortwave_mrt_delta[12] == pytest.approx(31.50184, rel=1e-3)
    assert isinstance(solarcal_obj.longwave_mrt_delta, HourlyContinuousCollection)
    assert len(solarcal_obj.longwave_mrt_delta.values) == calc_length
    assert solarcal_obj.longwave_mrt_delta[12] == pytest.approx(-6.614966, rel=1e-3)
    assert isinstance(solarcal_obj.mrt_delta, HourlyContinuousCollection)
    assert len(solarcal_obj.mrt_delta.values) == calc_length
    assert solarcal_obj.mrt_delta[12] == pytest.approx(24.886886, rel=1e-3)
    assert isinstance(solarcal_obj.mean_radiant_temperature, HourlyContinuousCollection)
    assert len(solarcal_obj.mean_radiant_temperature.values) == calc_length
    assert solarcal_obj.mean_radiant_temperature[12] == pytest.approx(48.88688, rel=1e-3)


def test_outdoor_solarcal_collection_defaults():
    """Test the default inputs assigned to the OutdoorSolarCal collection."""
    calc_length = 24
    irr_header = Header(Irradiance(), 'W/m2', AnalysisPeriod(end_month=1, end_day=1))
    dir_norm = HourlyContinuousCollection(irr_header, [500] * calc_length)
    diff_horiz = HourlyContinuousCollection(irr_header, [200] * calc_length)
    solarcal_obj = OutdoorSolarCal(Location(), dir_norm, diff_horiz, 350, 24)

    assert isinstance(solarcal_obj.fraction_body_exposed, HourlyContinuousCollection)
    assert len(solarcal_obj.fraction_body_exposed.values) == calc_length
    assert solarcal_obj.fraction_body_exposed[12] == 1

    assert isinstance(solarcal_obj.sky_exposure, HourlyContinuousCollection)
    assert len(solarcal_obj.sky_exposure.values) == calc_length
    assert solarcal_obj.sky_exposure[12] == 1

    assert isinstance(solarcal_obj.floor_reflectance, HourlyContinuousCollection)
    assert len(solarcal_obj.floor_reflectance.values) == calc_length
    assert solarcal_obj.floor_reflectance[12] == 0.25

    assert isinstance(solarcal_obj.solarcal_body_parameter, SolarCalParameter)
    default_par = SolarCalParameter()
    assert solarcal_obj.solarcal_body_parameter.posture == default_par.posture
    assert solarcal_obj.solarcal_body_parameter.sharp == default_par.sharp
    assert solarcal_obj.solarcal_body_parameter.body_azimuth == default_par.body_azimuth
    assert solarcal_obj.solarcal_body_parameter.body_absorptivity == default_par.body_absorptivity
    assert solarcal_obj.solarcal_body_parameter.body_emissivity == default_par.body_emissivity


def test_outdoor_solarcal_collection_full_input():
    """Test the initialization of the OutdoorSolarCal collection will all inputs."""
    calc_length = 24
    irr_header = Header(Irradiance(), 'W/m2', AnalysisPeriod(end_month=1, end_day=1))
    dir_norm = HourlyContinuousCollection(irr_header, [500] * calc_length)
    diff_horiz = HourlyContinuousCollection(irr_header, [200] * calc_length)
    custom_par = SolarCalParameter('seated', None, 45, 0.65, 0.97)
    solarcal_obj = OutdoorSolarCal(Location(), dir_norm, diff_horiz, 350, 24,
                                   0.6, 0.4, 0.35, custom_par)

    assert solarcal_obj.fraction_body_exposed[12] == 0.6
    assert solarcal_obj.sky_exposure[12] == 0.4
    assert solarcal_obj.floor_reflectance[0] == 0.35
    assert solarcal_obj.solarcal_body_parameter.posture == 'seated'
    assert solarcal_obj.solarcal_body_parameter.sharp is None
    assert solarcal_obj.solarcal_body_parameter.body_azimuth == 45
    assert solarcal_obj.solarcal_body_parameter.body_absorptivity == 0.65
    assert solarcal_obj.solarcal_body_parameter.body_emissivity == 0.97


def test_init_outdoor_solarcal_collection_epw():
    """Test the initialization of the OutdoorSolarCal collection with EPW input."""
    calc_length = 8760
    relative_path = './tests/epw/chicago.epw'
    epw = EPW(relative_path)
    solarcal_obj = OutdoorSolarCal(epw.location, epw.direct_normal_radiation,
                                   epw.diffuse_horizontal_radiation,
                                   epw.horizontal_infrared_radiation_intensity,
                                   epw.dry_bulb_temperature)

    assert isinstance(solarcal_obj.direct_normal_solar, HourlyContinuousCollection)
    assert len(solarcal_obj.direct_normal_solar.values) == calc_length
    assert solarcal_obj.direct_normal_solar[12] == pytest.approx(151, rel=1e-3)
    assert isinstance(solarcal_obj.diffuse_horizontal_solar, HourlyContinuousCollection)
    assert len(solarcal_obj.diffuse_horizontal_solar.values) == calc_length
    assert solarcal_obj.diffuse_horizontal_solar[12] == pytest.approx(168, rel=1e-3)

    assert isinstance(solarcal_obj.shortwave_effective_radiant_field, HourlyContinuousCollection)
    assert len(solarcal_obj.shortwave_effective_radiant_field.values) == calc_length
    assert solarcal_obj.shortwave_effective_radiant_field[12] == pytest.approx(83.29130, rel=1e-3)
    assert isinstance(solarcal_obj.longwave_effective_radiant_field, HourlyContinuousCollection)
    assert len(solarcal_obj.longwave_effective_radiant_field.values) == calc_length
    assert solarcal_obj.longwave_effective_radiant_field[12] == pytest.approx(-27.39307, rel=1e-3)
    assert isinstance(solarcal_obj.shortwave_mrt_delta, HourlyContinuousCollection)
    assert len(solarcal_obj.shortwave_mrt_delta.values) == calc_length
    assert solarcal_obj.shortwave_mrt_delta[12] == pytest.approx(19.109207, rel=1e-3)
    assert isinstance(solarcal_obj.longwave_mrt_delta, HourlyContinuousCollection)
    assert len(solarcal_obj.longwave_mrt_delta.values) == calc_length
    assert solarcal_obj.longwave_mrt_delta[12] == pytest.approx(-6.284688, rel=1e-3)
    assert isinstance(solarcal_obj.mrt_delta, HourlyContinuousCollection)
    assert len(solarcal_obj.mrt_delta.values) == calc_length
    assert solarcal_obj.mrt_delta[12] == pytest.approx(12.8245189, rel=1e-3)
    assert isinstance(solarcal_obj.mean_radiant_temperature, HourlyContinuousCollection)
    assert len(solarcal_obj.mean_radiant_temperature.values) == calc_length
    assert solarcal_obj.mean_radiant_temperature[12] == pytest.approx(9.524518, rel=1e-3)


def test_init_indoor_solarcal_collection():
    """Test the initialization of the IndoorSolarCal collection."""
    calc_length = 24
    irr_header = Header(Irradiance(), 'W/m2', AnalysisPeriod(end_month=1, end_day=1))
    dir_norm = HourlyContinuousCollection(irr_header, [500] * calc_length)
    diff_horiz = HourlyContinuousCollection(irr_header, [200] * calc_length)
    solarcal_obj = IndoorSolarCal(Location(), dir_norm, diff_horiz, 24)

    assert solarcal_obj.comfort_model == 'Indoor SolarCal'
    assert solarcal_obj.calc_length == calc_length
    str(solarcal_obj)  # test that the string representaiton is ok

    assert isinstance(solarcal_obj.direct_normal_solar, HourlyContinuousCollection)
    assert len(solarcal_obj.direct_normal_solar.values) == calc_length
    assert solarcal_obj.direct_normal_solar[12] == 500
    assert isinstance(solarcal_obj.diffuse_horizontal_solar, HourlyContinuousCollection)
    assert len(solarcal_obj.diffuse_horizontal_solar.values) == calc_length
    assert solarcal_obj.diffuse_horizontal_solar[12] == 200

    assert isinstance(solarcal_obj.effective_radiant_field, HourlyContinuousCollection)
    assert len(solarcal_obj.effective_radiant_field.values) == calc_length
    assert solarcal_obj.effective_radiant_field[12] == pytest.approx(54.9228, rel=1e-3)
    assert isinstance(solarcal_obj.mrt_delta, HourlyContinuousCollection)
    assert len(solarcal_obj.mrt_delta.values) == calc_length
    assert solarcal_obj.mrt_delta[12] == pytest.approx(12.600738, rel=1e-3)
    assert isinstance(solarcal_obj.mean_radiant_temperature, HourlyContinuousCollection)
    assert len(solarcal_obj.mean_radiant_temperature.values) == calc_length
    assert solarcal_obj.mean_radiant_temperature[12] == pytest.approx(36.600738, rel=1e-3)


def test_indoor_solarcal_collection_defaults():
    """Test the default inputs assigned to the IndoorSolarCal collection."""
    calc_length = 24
    irr_header = Header(Irradiance(), 'W/m2', AnalysisPeriod(end_month=1, end_day=1))
    dir_norm = HourlyContinuousCollection(irr_header, [500] * calc_length)
    diff_horiz = HourlyContinuousCollection(irr_header, [200] * calc_length)
    solarcal_obj = IndoorSolarCal(Location(), dir_norm, diff_horiz, 24)

    assert isinstance(solarcal_obj.fraction_body_exposed, HourlyContinuousCollection)
    assert len(solarcal_obj.fraction_body_exposed.values) == calc_length
    assert solarcal_obj.fraction_body_exposed[12] == 1

    assert isinstance(solarcal_obj.sky_exposure, HourlyContinuousCollection)
    assert len(solarcal_obj.sky_exposure.values) == calc_length
    assert solarcal_obj.sky_exposure[12] == 1

    assert isinstance(solarcal_obj.floor_reflectance, HourlyContinuousCollection)
    assert len(solarcal_obj.floor_reflectance.values) == calc_length
    assert solarcal_obj.floor_reflectance[12] == 0.25

    assert isinstance(solarcal_obj.window_transmittance, HourlyContinuousCollection)
    assert len(solarcal_obj.window_transmittance.values) == calc_length
    assert solarcal_obj.window_transmittance[12] == 0.4

    assert isinstance(solarcal_obj.solarcal_body_parameter, SolarCalParameter)
    default_par = SolarCalParameter()
    assert solarcal_obj.solarcal_body_parameter.posture == default_par.posture
    assert solarcal_obj.solarcal_body_parameter.sharp == default_par.sharp
    assert solarcal_obj.solarcal_body_parameter.body_azimuth == default_par.body_azimuth
    assert solarcal_obj.solarcal_body_parameter.body_absorptivity == default_par.body_absorptivity
    assert solarcal_obj.solarcal_body_parameter.body_emissivity == default_par.body_emissivity


def test_indoor_solarcal_collection_full_input():
    """Test the initialization of the IndoorSolarCal collection will all inputs."""
    calc_length = 24
    irr_header = Header(Irradiance(), 'W/m2', AnalysisPeriod(end_month=1, end_day=1))
    dir_norm = HourlyContinuousCollection(irr_header, [500] * calc_length)
    diff_horiz = HourlyContinuousCollection(irr_header, [200] * calc_length)
    custom_par = SolarCalParameter('seated', None, 45, 0.65, 0.97)
    solarcal_obj = IndoorSolarCal(Location(), dir_norm, diff_horiz, 24,
                                  0.6, 0.4, 0.35, 0.7, custom_par)

    assert solarcal_obj.fraction_body_exposed[12] == 0.6
    assert solarcal_obj.sky_exposure[12] == 0.4
    assert solarcal_obj.floor_reflectance[0] == 0.35
    assert solarcal_obj.window_transmittance[0] == 0.7
    assert solarcal_obj.solarcal_body_parameter.posture == 'seated'
    assert solarcal_obj.solarcal_body_parameter.sharp is None
    assert solarcal_obj.solarcal_body_parameter.body_azimuth == 45
    assert solarcal_obj.solarcal_body_parameter.body_absorptivity == 0.65
    assert solarcal_obj.solarcal_body_parameter.body_emissivity == 0.97


def test_init_indoor_solarcal_collection_epw():
    """Test the initialization of the IndoorSolarCal collection with EPW input."""
    calc_length = 8760
    relative_path = './tests/epw/chicago.epw'
    epw = EPW(relative_path)
    solarcal_obj = IndoorSolarCal(epw.location, epw.direct_normal_radiation,
                                  epw.diffuse_horizontal_radiation, 24)

    assert isinstance(solarcal_obj.direct_normal_solar, HourlyContinuousCollection)
    assert len(solarcal_obj.direct_normal_solar.values) == calc_length
    assert solarcal_obj.direct_normal_solar[12] == pytest.approx(151, rel=1e-3)
    assert isinstance(solarcal_obj.diffuse_horizontal_solar, HourlyContinuousCollection)
    assert len(solarcal_obj.diffuse_horizontal_solar.values) == calc_length
    assert solarcal_obj.diffuse_horizontal_solar[12] == pytest.approx(168, rel=1e-3)

    assert isinstance(solarcal_obj.effective_radiant_field, HourlyContinuousCollection)
    assert len(solarcal_obj.effective_radiant_field.values) == calc_length
    assert solarcal_obj.effective_radiant_field[12] == pytest.approx(33.31652, rel=1e-3)
    assert isinstance(solarcal_obj.mrt_delta, HourlyContinuousCollection)
    assert len(solarcal_obj.mrt_delta.values) == calc_length
    assert solarcal_obj.mrt_delta[12] == pytest.approx(7.6436828, rel=1e-3)
    assert isinstance(solarcal_obj.mean_radiant_temperature, HourlyContinuousCollection)
    assert len(solarcal_obj.mean_radiant_temperature.values) == calc_length
    assert solarcal_obj.mean_radiant_temperature[12] == pytest.approx(31.6436828, rel=1e-3)


def test_init_horizontal_solarcal_collection():
    """Test the initialization of the HorizontalSolarCal collection."""
    calc_length = 24
    irr_header = Header(Irradiance(), 'W/m2', AnalysisPeriod(end_month=1, end_day=1))
    dir_norm = HourlyContinuousCollection(irr_header, [300] * calc_length)
    diff_horiz = HourlyContinuousCollection(irr_header, [100] * calc_length)
    solarcal_obj = HorizontalSolarCal(Location(), dir_norm, diff_horiz, 24)

    assert solarcal_obj.comfort_model == 'Horizontal SolarCal'
    assert solarcal_obj.calc_length == calc_length
    str(solarcal_obj)  # test that the string representaiton is ok

    assert isinstance(solarcal_obj.direct_horizontal_solar, HourlyContinuousCollection)
    assert len(solarcal_obj.direct_horizontal_solar.values) == calc_length
    assert solarcal_obj.direct_horizontal_solar[12] == 300
    assert isinstance(solarcal_obj.diffuse_horizontal_solar, HourlyContinuousCollection)
    assert len(solarcal_obj.diffuse_horizontal_solar.values) == calc_length
    assert solarcal_obj.diffuse_horizontal_solar[12] == 100

    assert isinstance(solarcal_obj.effective_radiant_field, HourlyContinuousCollection)
    assert len(solarcal_obj.effective_radiant_field.values) == calc_length
    assert solarcal_obj.effective_radiant_field[12] == pytest.approx(79.35027, rel=1e-3)
    assert isinstance(solarcal_obj.mrt_delta, HourlyContinuousCollection)
    assert len(solarcal_obj.mrt_delta.values) == calc_length
    assert solarcal_obj.mrt_delta[12] == pytest.approx(18.20503, rel=1e-3)
    assert isinstance(solarcal_obj.mean_radiant_temperature, HourlyContinuousCollection)
    assert len(solarcal_obj.mean_radiant_temperature.values) == calc_length
    assert solarcal_obj.mean_radiant_temperature[12] == pytest.approx(42.20503, rel=1e-3)


def test_horizontal_solarcal_collection_defaults():
    """Test the default inputs assigned to the HorizontalSolarCal collection."""
    calc_length = 24
    irr_header = Header(Irradiance(), 'W/m2', AnalysisPeriod(end_month=1, end_day=1))
    dir_norm = HourlyContinuousCollection(irr_header, [500] * calc_length)
    diff_horiz = HourlyContinuousCollection(irr_header, [200] * calc_length)
    solarcal_obj = HorizontalSolarCal(Location(), dir_norm, diff_horiz, 24)

    assert isinstance(solarcal_obj.fraction_body_exposed, HourlyContinuousCollection)
    assert len(solarcal_obj.fraction_body_exposed.values) == calc_length
    assert solarcal_obj.fraction_body_exposed[12] == 1

    assert isinstance(solarcal_obj.floor_reflectance, HourlyContinuousCollection)
    assert len(solarcal_obj.floor_reflectance.values) == calc_length
    assert solarcal_obj.floor_reflectance[12] == 0.25

    assert isinstance(solarcal_obj.solarcal_body_parameter, SolarCalParameter)
    default_par = SolarCalParameter()
    assert solarcal_obj.solarcal_body_parameter.posture == default_par.posture
    assert solarcal_obj.solarcal_body_parameter.sharp == default_par.sharp
    assert solarcal_obj.solarcal_body_parameter.body_azimuth == default_par.body_azimuth
    assert solarcal_obj.solarcal_body_parameter.body_absorptivity == default_par.body_absorptivity
    assert solarcal_obj.solarcal_body_parameter.body_emissivity == default_par.body_emissivity


def test_horizontal_solarcal_collection_full_input():
    """Test the initialization of the HorizontalSolarCal collection will all inputs."""
    calc_length = 24
    irr_header = Header(Irradiance(), 'W/m2', AnalysisPeriod(end_month=1, end_day=1))
    dir_norm = HourlyContinuousCollection(irr_header, [500] * calc_length)
    diff_horiz = HourlyContinuousCollection(irr_header, [200] * calc_length)
    custom_par = SolarCalParameter('seated', None, 45, 0.65, 0.97)
    solarcal_obj = HorizontalSolarCal(Location(), dir_norm, diff_horiz, 24,
                                      0.6, 0.35, custom_par)

    assert solarcal_obj.fraction_body_exposed[12] == 0.6
    assert solarcal_obj.floor_reflectance[0] == 0.35
    assert solarcal_obj.solarcal_body_parameter.posture == 'seated'
    assert solarcal_obj.solarcal_body_parameter.sharp is None
    assert solarcal_obj.solarcal_body_parameter.body_azimuth == 45
    assert solarcal_obj.solarcal_body_parameter.body_absorptivity == 0.65
    assert solarcal_obj.solarcal_body_parameter.body_emissivity == 0.97


def test_init_horizontal_solarcal_collection_epw():
    """Test the initialization of the HorizontalSolarCal collection with EPW input."""
    calc_length = 8760
    wea_obj = Wea.from_epw_file('./tests/epw/chicago.epw')
    diff_hr = wea_obj.diffuse_horizontal_irradiance
    dir_hr = wea_obj.direct_horizontal_irradiance
    solarcal_obj = HorizontalSolarCal(wea_obj.location, dir_hr, diff_hr, 24)

    assert isinstance(solarcal_obj.direct_horizontal_solar, HourlyContinuousCollection)
    assert len(solarcal_obj.direct_horizontal_solar.values) == calc_length
    assert solarcal_obj.direct_horizontal_solar[12] == pytest.approx(62.9354, rel=1e-3)
    assert isinstance(solarcal_obj.diffuse_horizontal_solar, HourlyContinuousCollection)
    assert len(solarcal_obj.diffuse_horizontal_solar.values) == calc_length
    assert solarcal_obj.diffuse_horizontal_solar[12] == pytest.approx(168, rel=1e-3)

    assert isinstance(solarcal_obj.effective_radiant_field, HourlyContinuousCollection)
    assert len(solarcal_obj.effective_radiant_field.values) == calc_length
    assert solarcal_obj.effective_radiant_field[12] == pytest.approx(82.80573, rel=1e-3)
    assert isinstance(solarcal_obj.mrt_delta, HourlyContinuousCollection)
    assert len(solarcal_obj.mrt_delta.values) == calc_length
    assert solarcal_obj.mrt_delta[12] == pytest.approx(18.997806, rel=1e-3)
    assert isinstance(solarcal_obj.mean_radiant_temperature, HourlyContinuousCollection)
    assert len(solarcal_obj.mean_radiant_temperature.values) == calc_length
    assert solarcal_obj.mean_radiant_temperature[12] == pytest.approx(42.997806, rel=1e-3)



# coding=utf-8
"""Parameters for specifying acceptable thermal conditions using the PMV model."""
from __future__ import division

from ._base import ComfortParameter
from ..pmv import ppd_threshold_from_comfort_class


class PMVParameter(ComfortParameter):
    """Parameters of PMV comfort.

    Properties:
        ppd_comfort_thresh
        humid_ratio_upper
        humid_ratio_low
        still_air_threshold
    """
    _model = 'Predicted Mean Vote'

    def __init__(self, ppd_comfort_thresh=None, humid_ratio_upper=None,
                 humid_ratio_lower=None, still_air_threshold=None):
        """Initalize PMV Parameters.

        Args:
            ppd_comfort_thresh: A number between 5 and 100 that represents the upper
                threshold of PPD that is considered acceptable.
                Default is 10, which charcterizes most buildings in the ASHRAE-55 and
                EN-15251 standards.
            humid_ratio_upper: A number between 0 and 1 indicating the upper limit of
                humidity ratio that is considered acceptable. Default is 1 for
                essentially no limit.
            humid_ratio_lower: A number between 0 and 1 indicating the lower limit of
                humidity ratio considered acceptable. Default is 0 for essentially
                no limit.
            still_air_threshold: The air speed threshold in m/s at which the standard
                effective temperature (SET) model will be used to correct for the
                cooling effect of elevated air speeds.
                Default is 0.1 m/s, which is the limit according to ASHRAE-55.
        """

        self.ppd_comfort_thresh = \
            ppd_comfort_thresh if ppd_comfort_thresh is not None else 10
        self._hr_upper = humid_ratio_upper if humid_ratio_upper is not None else 1
        self._hr_lower = humid_ratio_lower if humid_ratio_lower is not None else 0
        self._still_thresh = \
            still_air_threshold if still_air_threshold is not None else 0.1

        assert 0 <= self._hr_upper <= 1, \
            'humid_ratio_upper must be between 0 and 1. Got {}'.format(
                self._hr_upper)
        assert 0 <= self._hr_lower <= 1, \
            'humid_ratio_lower must be between 0 and 1. Got {}'.format(
                self._hr_lower)
        assert self._hr_lower <= self._hr_upper, \
            'humid_ratio_lower must be less than humid_ratio_upper. {} > {}'.format(
                self._hr_lower, self._hr_upper)

        assert 0 <= self._still_thresh, \
            'still_air_threshold must be greater than 0. Got {}'.format(
                self._still_thresh)

    @property
    def ppd_comfort_thresh(self):
        """The threshold of the percentage of people dissatisfied (PPD) beyond which
        the conditions are not acceptable.  The default is 10%.
        """
        return self._ppd_thresh

    @ppd_comfort_thresh.setter
    def ppd_comfort_thresh(self, ppd):
        assert 5 <= ppd <= 100, \
            'ppd_comfort_thresh must be between 5 and 100. Got {}'.format(ppd)
        self._ppd_thresh = ppd

    @property
    def humid_ratio_upper(self):
        """A number representing the upper boundary of humidity ratio above which conditions
        are considered too humid to be comfortable.  The default is set to 0.03 kg
        wather/kg air.
        """
        return self._hr_upper

    @property
    def humid_ratio_lower(self):
        """A number representing the lower boundary of humidity ratio below which conditions
        are considered too dry to be comfortable.  The default is set to 0 kg wather/kg
        air.
        """
        return self._hr_lower

    @property
    def still_air_threshold(self):
        """A number representing the wind speed beyond which the formula for Standard
        Effective Temperature (SET) is used to dtermine PMV/PPD (as opposed to Fanger's
        original equation). The default is set to 0.1 m/s.
        """
        return self._still_thresh

    def set_ppd_comfort_thresh_from_comfort_class(self, comfort_class):
        """Set the PPD threshold given the EN-15251 comfort class."""
        self.ppd_comfort_thresh = ppd_threshold_from_comfort_class(comfort_class)

    def is_comfortable(self, ppd, humidity_ratio=0):
        """Determine if conditions are comfortable or not.

        Values are one of the following:
            0 = uncomfortable
            1 = comfortable
        """
        return 1 if (ppd <= self._ppd_thresh and humidity_ratio >= self._hr_lower and
                     humidity_ratio <= self._hr_upper) else 0

    def thermal_condition(self, pmv, ppd):
        """Determine whether conditions are cold, neutral or hot.

        Values are one of the following:
            -1 = cold
             0 = netural
            +1 = hot
        """
        if ppd >= self._ppd_thresh:
            return 1 if pmv > 0 else -1
        else:
            return 0

    def discomfort_reason(self, pmv, ppd, humidity_ratio=0):
        """Determine the reason why conditions are comfortable or not.

        Values are one of the following:
            -2 = too dry
            -1 = too cold
             0 = comfortable
            +1 = too hot
            +2 = too humid
        """
        if ppd >= self._ppd_thresh:
            return 1 if pmv > 0 else -1
        elif humidity_ratio < self._hr_lower:
            return -2
        elif humidity_ratio > self._hr_upper:
            return 2
        else:
            return 0

    def duplicate(self):
        """Duplicate these comfort parameters."""
        return PMVParameter(self.ppd_comfort_thresh, self.humid_ratio_upper,
                            self.humid_ratio_lower, self.still_air_threshold)

    def __repr__(self):
        """PMV comfort parameters representation."""
        return "PMV Comfort Parameters\n PPD Threshold: {}\n HR Upper: {}"\
            "\n HR Lower: {}\n Still Air Threshold: {}".format(
                self._ppd_thresh, self._hr_upper, self._hr_lower, self._still_thresh)



# coding utf-8
import pytest

from ladybug_comfort.collection.utci import UTCI
from ladybug_comfort.parameter.utci import UTCIParameter

from ladybug_comfort.utci import universal_thermal_climate_index, calc_missing_utci_input

from ladybug.analysisperiod import AnalysisPeriod
from ladybug.header import Header
from ladybug.datacollection import HourlyContinuousCollection
from ladybug.epw import EPW

from ladybug.datatype.temperature import Temperature
from ladybug.datatype.fraction import RelativeHumidity
from ladybug.datatype.speed import AirSpeed


def test_utci():
    """Test the utci function"""
    assert universal_thermal_climate_index(20, 20, 3, 50) == \
        pytest.approx(16.24224, rel=1e-2)
    assert universal_thermal_climate_index(30, 30, 0.5, 90) == \
        pytest.approx(35.511294, rel=1e-2)


def test_calc_missing_utci_input():
    """Test the calc_missing_utci_input function"""
    input_1 = {'ta': None, 'tr': 20, 'vel': 0.5, 'rh': 50}
    input_2 = {'ta': 20, 'tr': None, 'vel': 0.5, 'rh': 50}
    input_3 = {'ta': 22, 'tr': 22, 'vel': None, 'rh': 50}
    input_4 = {'ta': 20, 'tr': 20, 'vel': 0.5, 'rh': None}
    input_5 = {'ta': None, 'tr': None, 'vel': 5, 'rh': 50}
    updated_input_1 = calc_missing_utci_input(25, input_1)
    updated_input_2 = calc_missing_utci_input(25, input_2)
    updated_input_3 = calc_missing_utci_input(15, input_3, up_bound=1)
    updated_input_4 = calc_missing_utci_input(22, input_4)
    updated_input_5 = calc_missing_utci_input(22, input_5)
    assert updated_input_1['ta'] == pytest.approx(26.9827, rel=1e-2)
    assert updated_input_2['tr'] == pytest.approx(36.3803, rel=1e-2)
    assert updated_input_3['vel'] == pytest.approx(5.77514, rel=1e-2)
    assert updated_input_4['rh'] == pytest.approx(90.388989, rel=1e-2)
    assert updated_input_5['ta'] == pytest.approx(26.413594, rel=1e-2)
    assert updated_input_5['ta'] == updated_input_5['tr']


def test_utci_parameters():
    """Test UTCI Parameters."""
    cold_thresh = 8
    heat_thresh = 27
    extreme_cold_thresh = -41
    very_strong_cold_thresh = -28
    strong_cold_thresh = -14
    moderate_cold_thresh = -1
    moderate_heat_thresh = 29
    strong_heat_thresh = 33
    very_strong_heat_thresh = 39
    extreme_heat_thresh = 47

    utci_comf = UTCIParameter(
        cold_thresh, heat_thresh, extreme_cold_thresh, very_strong_cold_thresh,
        strong_cold_thresh, moderate_cold_thresh, moderate_heat_thresh,
        strong_heat_thresh, very_strong_heat_thresh, extreme_heat_thresh)

    assert utci_comf.cold_thresh == cold_thresh
    assert utci_comf.heat_thresh == heat_thresh
    assert utci_comf.extreme_cold_thresh == extreme_cold_thresh
    assert utci_comf.very_strong_cold_thresh == very_strong_cold_thresh
    assert utci_comf.strong_cold_thresh == strong_cold_thresh
    assert utci_comf.moderate_cold_thresh == moderate_cold_thresh
    assert utci_comf.moderate_heat_thresh == moderate_heat_thresh
    assert utci_comf.strong_heat_thresh == strong_heat_thresh
    assert utci_comf.very_strong_heat_thresh == very_strong_heat_thresh
    assert utci_comf.extreme_heat_thresh == extreme_heat_thresh


def test_utci_parameters_invalid():
    """Test UTCI Parameters for invalid inputs."""
    cold_thresh = 30
    heat_thresh = 8

    with pytest.raises(AssertionError):
        UTCIParameter(cold_thresh=cold_thresh)
    with pytest.raises(AssertionError):
        UTCIParameter(heat_thresh=heat_thresh)


def test_comfort_check():
    """Test comfort check on UTCI Parameters."""
    utci_comf = UTCIParameter()
    comf_test = utci_comf.is_comfortable(5)
    assert comf_test == 0
    comf_test = utci_comf.is_comfortable(22)
    assert comf_test == 1


def test_thermal_condition_check():
    """Test the thermal condition check on UTCI Parameters."""
    utci_comf = UTCIParameter()
    condition_test = utci_comf.thermal_condition(5)
    assert condition_test == -1
    condition_test = utci_comf.thermal_condition(22)
    assert condition_test == 0
    condition_test = utci_comf.thermal_condition(32)
    assert condition_test == 1


def test_thermal_condition_five_point_check():
    """Test the thermal condition check on UTCI Parameters."""
    utci_comf = UTCIParameter()
    condition_test = utci_comf.thermal_condition_five_point(-15)
    assert condition_test == -2
    condition_test = utci_comf.thermal_condition_five_point(5)
    assert condition_test == -1
    condition_test = utci_comf.thermal_condition_five_point(22)
    assert condition_test == 0
    condition_test = utci_comf.thermal_condition_five_point(30)
    assert condition_test == 1
    condition_test = utci_comf.thermal_condition_five_point(36)
    assert condition_test == 2


def test_thermal_condition_seven_point_check():
    """Test the thermal condition check on UTCI Parameters."""
    utci_comf = UTCIParameter()
    condition_test = utci_comf.thermal_condition_seven_point(-30)
    assert condition_test == -3
    condition_test = utci_comf.thermal_condition_seven_point(-15)
    assert condition_test == -2
    condition_test = utci_comf.thermal_condition_seven_point(5)
    assert condition_test == -1
    condition_test = utci_comf.thermal_condition_seven_point(22)
    assert condition_test == 0
    condition_test = utci_comf.thermal_condition_seven_point(30)
    assert condition_test == 1
    condition_test = utci_comf.thermal_condition_seven_point(36)
    assert condition_test == 2
    condition_test = utci_comf.thermal_condition_seven_point(40)
    assert condition_test == 3


def test_thermal_condition_none_point_check():
    """Test the thermal condition check on UTCI Parameters."""
    utci_comf = UTCIParameter()
    condition_test = utci_comf.thermal_condition_nine_point(-30)
    assert condition_test == -4
    condition_test = utci_comf.thermal_condition_nine_point(-18)
    assert condition_test == -3
    condition_test = utci_comf.thermal_condition_nine_point(-8)
    assert condition_test == -2
    condition_test = utci_comf.thermal_condition_nine_point(5)
    assert condition_test == -1
    condition_test = utci_comf.thermal_condition_nine_point(22)
    assert condition_test == 0
    condition_test = utci_comf.thermal_condition_nine_point(27)
    assert condition_test == 1
    condition_test = utci_comf.thermal_condition_nine_point(30)
    assert condition_test == 2
    condition_test = utci_comf.thermal_condition_nine_point(36)
    assert condition_test == 3
    condition_test = utci_comf.thermal_condition_nine_point(40)
    assert condition_test == 4


def test_thermal_condition_eleven_point_check():
    """Test the thermal condition check on UTCI Parameters."""
    utci_comf = UTCIParameter()
    condition_test = utci_comf.thermal_condition_eleven_point(-50)
    assert condition_test == -5
    condition_test = utci_comf.thermal_condition_eleven_point(-30)
    assert condition_test == -4
    condition_test = utci_comf.thermal_condition_eleven_point(-18)
    assert condition_test == -3
    condition_test = utci_comf.thermal_condition_eleven_point(-8)
    assert condition_test == -2
    condition_test = utci_comf.thermal_condition_eleven_point(5)
    assert condition_test == -1
    condition_test = utci_comf.thermal_condition_eleven_point(22)
    assert condition_test == 0
    condition_test = utci_comf.thermal_condition_eleven_point(27)
    assert condition_test == 1
    condition_test = utci_comf.thermal_condition_eleven_point(30)
    assert condition_test == 2
    condition_test = utci_comf.thermal_condition_eleven_point(36)
    assert condition_test == 3
    condition_test = utci_comf.thermal_condition_eleven_point(40)
    assert condition_test == 4
    condition_test = utci_comf.thermal_condition_eleven_point(50)
    assert condition_test == 5


def test_original_utci_category():
    """Test the thermal condition check on UTCI Parameters."""
    utci_comf = UTCIParameter()
    condition_test = utci_comf.original_utci_category(-50)
    assert condition_test == 0
    condition_test = utci_comf.original_utci_category(-30)
    assert condition_test == 1
    condition_test = utci_comf.original_utci_category(-18)
    assert condition_test == 2
    condition_test = utci_comf.original_utci_category(-8)
    assert condition_test == 3
    condition_test = utci_comf.original_utci_category(5)
    assert condition_test == 4
    condition_test = utci_comf.original_utci_category(22)
    assert condition_test == 5
    condition_test = utci_comf.original_utci_category(30)
    assert condition_test == 6
    condition_test = utci_comf.original_utci_category(36)
    assert condition_test == 7
    condition_test = utci_comf.original_utci_category(40)
    assert condition_test == 8
    condition_test = utci_comf.original_utci_category(50)
    assert condition_test == 9


def test_init_utci_collection():
    """Test the initialization of the UTCI collection and basic outputs."""
    calc_length = 24
    air_temp_header = Header(Temperature(), 'C', AnalysisPeriod(end_month=1, end_day=1))
    air_temp = HourlyContinuousCollection(air_temp_header, [24] * calc_length)
    utci_obj = UTCI(air_temp, 50)

    assert utci_obj.comfort_model == 'Universal Thermal Climate Index'
    assert utci_obj.calc_length == calc_length
    str(utci_obj)  # test that the string representaiton is ok

    assert isinstance(utci_obj.air_temperature, HourlyContinuousCollection)
    assert len(utci_obj.air_temperature.values) == calc_length
    assert utci_obj.air_temperature[0] == 24
    assert isinstance(utci_obj.rel_humidity, HourlyContinuousCollection)
    assert len(utci_obj.rel_humidity.values) == calc_length
    assert utci_obj.rel_humidity[0] == 50

    assert isinstance(utci_obj.universal_thermal_climate_index, HourlyContinuousCollection)
    assert len(utci_obj.universal_thermal_climate_index.values) == calc_length
    assert utci_obj.universal_thermal_climate_index[0] == pytest.approx(23.8110341, rel=1e-3)


def test_utci_collection_defaults():
    """Test the default inputs assigned to the UTCI collection."""
    calc_length = 24
    air_temp_header = Header(Temperature(), 'C', AnalysisPeriod(end_month=1, end_day=1))
    air_temp = HourlyContinuousCollection(air_temp_header, [24] * calc_length)
    utci_obj = UTCI(air_temp, 50)

    assert isinstance(utci_obj.rad_temperature, HourlyContinuousCollection)
    assert len(utci_obj.rad_temperature.values) == calc_length
    assert utci_obj.rad_temperature[0] == utci_obj.air_temperature[0]

    assert isinstance(utci_obj.wind_speed, HourlyContinuousCollection)
    assert len(utci_obj.wind_speed.values) == calc_length
    assert utci_obj.wind_speed[0] == 0.1

    assert isinstance(utci_obj.comfort_parameter, UTCIParameter)
    default_par = UTCIParameter()
    assert utci_obj.comfort_parameter.cold_thresh == default_par.cold_thresh
    assert utci_obj.comfort_parameter.heat_thresh == default_par.heat_thresh
    assert utci_obj.comfort_parameter.extreme_cold_thresh == default_par.extreme_cold_thresh
    assert utci_obj.comfort_parameter.very_strong_cold_thresh == default_par.very_strong_cold_thresh
    assert utci_obj.comfort_parameter.strong_cold_thresh == default_par.strong_cold_thresh
    assert utci_obj.comfort_parameter.moderate_cold_thresh == default_par.moderate_cold_thresh
    assert utci_obj.comfort_parameter.moderate_heat_thresh == default_par.moderate_heat_thresh
    assert utci_obj.comfort_parameter.strong_heat_thresh == default_par.strong_heat_thresh
    assert utci_obj.comfort_parameter.very_strong_heat_thresh == default_par.very_strong_heat_thresh
    assert utci_obj.comfort_parameter.extreme_heat_thresh == default_par.extreme_heat_thresh


def test_utci_collection_comfort_outputs():
    """Test the is_comfortable and thermal_condition outputs of the UTCI collection."""
    calc_length = 24
    air_temp_header = Header(Temperature(), 'C', AnalysisPeriod(end_month=1, end_day=1))
    air_temp = HourlyContinuousCollection(air_temp_header, range(5, 5 + calc_length))
    utci_obj = UTCI(air_temp, 50)

    assert isinstance(utci_obj.is_comfortable, HourlyContinuousCollection)
    assert len(utci_obj.is_comfortable.values) == calc_length
    assert utci_obj.is_comfortable[0] == 0
    assert utci_obj.is_comfortable[12] == 1
    assert utci_obj.is_comfortable[23] == 0

    assert isinstance(utci_obj.thermal_condition, HourlyContinuousCollection)
    assert len(utci_obj.thermal_condition.values) == calc_length
    assert utci_obj.thermal_condition[0] == -1
    assert utci_obj.thermal_condition[12] == 0
    assert utci_obj.thermal_condition[23] == 1

    assert isinstance(utci_obj.thermal_condition_five_point, HourlyContinuousCollection)
    assert len(utci_obj.thermal_condition_five_point.values) == calc_length
    assert utci_obj.thermal_condition_five_point[0] == -1
    assert utci_obj.thermal_condition_five_point[12] == 0
    assert utci_obj.thermal_condition_five_point[23] == 1

    assert isinstance(utci_obj.thermal_condition_seven_point, HourlyContinuousCollection)
    assert len(utci_obj.thermal_condition_seven_point.values) == calc_length
    assert utci_obj.thermal_condition_seven_point[0] == -1
    assert utci_obj.thermal_condition_seven_point[12] == 0
    assert utci_obj.thermal_condition_seven_point[23] == 1

    assert isinstance(utci_obj.thermal_condition_nine_point, HourlyContinuousCollection)
    assert len(utci_obj.thermal_condition_nine_point.values) == calc_length
    assert utci_obj.thermal_condition_nine_point[0] == -1
    assert utci_obj.thermal_condition_nine_point[12] == 0
    assert utci_obj.thermal_condition_nine_point[22] == 1
    assert utci_obj.thermal_condition_nine_point[23] == 2

    assert isinstance(utci_obj.thermal_condition_eleven_point, HourlyContinuousCollection)
    assert len(utci_obj.thermal_condition_eleven_point.values) == calc_length
    assert utci_obj.thermal_condition_eleven_point[0] == -1
    assert utci_obj.thermal_condition_eleven_point[12] == 0
    assert utci_obj.thermal_condition_eleven_point[22] == 1
    assert utci_obj.thermal_condition_eleven_point[23] == 2

    assert isinstance(utci_obj.original_utci_category, HourlyContinuousCollection)
    assert len(utci_obj.original_utci_category.values) == calc_length
    assert utci_obj.original_utci_category[0] == 4
    assert utci_obj.original_utci_category[12] == 5
    assert utci_obj.original_utci_category[23] == 6


def test_utci_collection_immutability():
    """Test that the UTCI collection is immutable."""
    calc_length = 24
    air_temp_header = Header(Temperature(), 'C', AnalysisPeriod(end_month=1, end_day=1))
    air_temp = HourlyContinuousCollection(air_temp_header, [24] * calc_length)
    utci_obj = UTCI(air_temp, 50)

    # check that editing the original collection does not mutate the object
    air_temp[0] = 26
    assert utci_obj.air_temperature[0] == 24

    # check that editing collection properties does not mutate the object
    with pytest.raises(Exception):
        utci_obj.air_temperature[0] = 26
    with pytest.raises(Exception):
        utci_obj.air_temperature.values = [26] * calc_length
    with pytest.raises(Exception):
        utci_obj.air_temperature = air_temp
    with pytest.raises(Exception):
        utci_obj.universal_thermal_climate_index[0] = 20
    with pytest.raises(Exception):
        utci_obj.universal_thermal_climate_index.values = [20] * calc_length

    # check that properties cannot be edited directly
    with pytest.raises(Exception):
        utci_obj.universal_thermal_climate_index = air_temp
    with pytest.raises(Exception):
        utci_obj.comfort_parameter = UTCIParameter()


def test_init_utci_collection_full_input():
    """Test the initialization of the UTCI collection will all inputs."""
    calc_length = 24
    air_temp_header = Header(Temperature(), 'C', AnalysisPeriod(end_month=1, end_day=1))
    air_temp = HourlyContinuousCollection(air_temp_header, [24] * calc_length)
    custom_par = UTCIParameter(8, 25, -41, -28, -14, -1, 27, 31, 37, 45)
    utci_obj = UTCI(air_temp, 50, 22, 0.5, custom_par)

    assert utci_obj.air_temperature[0] == 24
    assert utci_obj.rel_humidity[0] == 50
    assert utci_obj.rad_temperature[0] == 22
    assert utci_obj.wind_speed[0] == 0.5
    assert utci_obj.comfort_parameter.cold_thresh == 8
    assert utci_obj.comfort_parameter.heat_thresh == 25
    assert utci_obj.comfort_parameter.extreme_cold_thresh == -41
    assert utci_obj.comfort_parameter.very_strong_cold_thresh == -28
    assert utci_obj.comfort_parameter.strong_cold_thresh == -14
    assert utci_obj.comfort_parameter.moderate_cold_thresh == -1
    assert utci_obj.comfort_parameter.moderate_heat_thresh == 27
    assert utci_obj.comfort_parameter.strong_heat_thresh == 31
    assert utci_obj.comfort_parameter.very_strong_heat_thresh == 37
    assert utci_obj.comfort_parameter.extreme_heat_thresh == 45


def test_init_utci_collection_full_collection_input():
    """Test initialization of the UTCI collection will all inputs as collections."""
    calc_length = 24
    air_temp_header = Header(Temperature(), 'C', AnalysisPeriod(end_month=1, end_day=1))
    air_temp = HourlyContinuousCollection(air_temp_header, [24] * calc_length)
    rel_humid_header = Header(RelativeHumidity(), '%', AnalysisPeriod(end_month=1, end_day=1))
    rel_humid = HourlyContinuousCollection(rel_humid_header, [50] * calc_length)
    rad_temp_header = Header(Temperature(), 'C', AnalysisPeriod(end_month=1, end_day=1))
    rad_temp = HourlyContinuousCollection(rad_temp_header, [22] * calc_length)
    wind_speed_header = Header(AirSpeed(), 'm/s', AnalysisPeriod(end_month=1, end_day=1))
    wind_speed = HourlyContinuousCollection(wind_speed_header, [0.5] * calc_length)

    utci_obj = UTCI(air_temp, rel_humid, rad_temp, wind_speed)

    assert utci_obj.air_temperature[0] == 24
    assert utci_obj.rel_humidity[0] == 50
    assert utci_obj.rad_temperature[0] == 22
    assert utci_obj.wind_speed[0] == 0.5


def test_init_utci_collection_epw():
    """Test the initialization of the UTCI collection with EPW input."""
    calc_length = 8760
    relative_path = './tests/epw/chicago.epw'
    epw = EPW(relative_path)
    utci_obj = UTCI(epw.dry_bulb_temperature, epw.relative_humidity)

    assert len(utci_obj.air_temperature.values) == calc_length
    assert utci_obj.air_temperature[0] == -6.1
    assert len(utci_obj.rel_humidity.values) == calc_length
    assert utci_obj.rel_humidity[0] == 81

    assert len(utci_obj.universal_thermal_climate_index.values) == calc_length
    assert utci_obj.universal_thermal_climate_index[0] == pytest.approx(-5.367017, rel=1e-3)
    assert len(utci_obj.thermal_condition_eleven_point.values) == calc_length
    assert utci_obj.thermal_condition_eleven_point[0] == -2


def test_utci_collection_comfort_percent_outputs():
    """Test the is_comfortable and percent outputs of the UTCI collection."""
    relative_path = './tests/epw/chicago.epw'
    epw = EPW(relative_path)
    utci_obj = UTCI(epw.dry_bulb_temperature, epw.relative_humidity, wind_speed=epw.wind_speed)

    assert utci_obj.percent_comfortable == pytest.approx(35.6849315, rel=1e-3)
    assert utci_obj.percent_uncomfortable == pytest.approx(64.31506849, rel=1e-3)
    assert utci_obj.percent_neutral == pytest.approx(35.6849315, rel=1e-3)
    assert utci_obj.percent_hot == pytest.approx(3.3447488, rel=1e-3)
    assert utci_obj.percent_cold == pytest.approx(60.970319, rel=1e-3)

    assert utci_obj.percent_extreme_cold_stress == pytest.approx(0.35388127, rel=1e-3)
    assert utci_obj.percent_very_strong_cold_stress == pytest.approx(4.37214611, rel=1e-3)
    assert utci_obj.percent_strong_cold_stress == pytest.approx(16.541095, rel=1e-3)
    assert utci_obj.percent_moderate_cold_stress == pytest.approx(23.8356164, rel=1e-3)
    assert utci_obj.percent_slight_cold_stress == pytest.approx(15.8675799, rel=1e-3)

    assert utci_obj.percent_slight_heat_stress == pytest.approx(1.826484, rel=1e-3)
    assert utci_obj.percent_moderate_heat_stress == pytest.approx(1.312785, rel=1e-3)
    assert utci_obj.percent_strong_heat_stress == pytest.approx(0.2054794, rel=1e-3)
    assert utci_obj.percent_very_strong_heat_stress == 0.0
    assert utci_obj.percent_extreme_heat_stress == 0.0



# coding=utf-8
"""Utility functions for calculating Heat Index (HI)."""
from __future__ import division

import math


def heat_index(ta, rh):
    """Calculate heat index (HI) from air temperature and relative humidity.

    Heat index is derived from original work carried out by Robert G. Steadman [1],
    which defined heat index through large tables of emprical data.
    The formula here approximates the heat index to within +/- 0.7C and is
    the result of a multivariate fit [2].
    Heat index was adopted by the US's National Weather Service (NWS) in 1979.

    Note:
        [1] Steadman, R. G. (July 1979). "The Assessment of Sultriness. Part I: A
        Temperature-Humidity Index Based on Human Physiology and Clothing Science".
        Journal of Applied Meteorology. 18 (7): 861–873.

        [2] Lans P. Rothfusz. "The Heat Index 'Equation' (or, More Than You Ever
        Wanted to Know About Heat Index)", Scientific Services Division
        (NWS Southern Region Headquarters), 1 July 1990.
        https://www.weather.gov/media/ffc/ta_htindx.PDF

    Args:
        ta: Air temperature [C]
        rh: Relative humidity [%]

    Returns:
        hi: Heat index [C]
    """
    tf = ta * 9. / 5. + 32.  # convert to farenheit

    if tf < 80:
        hif = 0.5 * (tf + 61.0 + ((tf - 68.0) * 1.2) + (rh * 0.094))
    else:
        hif = -42.379 + 2.04901523 * tf + \
            10.14333127 * rh - \
            0.22475541 * tf * rh - \
            6.83783e-3 * tf ** 2 - \
            5.481717e-2 * rh ** 2 + \
            1.22874e-3 * tf ** 2 * rh + \
            8.5282e-4 * tf * rh ** 2 - \
            1.99e-6 * tf ** 2 * rh ** 2
        if tf >= 80 and tf <= 112 and rh < 13:
            adjust = ((13. - rh) / 4.) * math.sqrt((17. - abs(tf - 95.)) / 17.)
            hif = hif - adjust
        elif tf >= 80 and tf <= 87 and rh > 85:
            adjust = ((rh - 85) / 10) * ((87 - tf) / 5)
            hif = hif + adjust

    hi = (hif - 32.) * 5. / 9.  # convert to celcius

    return hi


def heat_index_warning_category(hi):
    """Get the category of warning associated with a given heat index (HI).

    Categories are used by the US National Weather Service (NWS) and National
    Oceanic and Atmospheric Administration (NOAA) to issue the following warnings:
        0 = No Warning. Satisfactory temperature. Can continue with activity.
        1 = Caution: Fatigue is possible with prolonged exposure and activity.
            Continuing activity could result in heat cramps.
        2 = Extreme caution: Heat cramps and heat exhaustion are possible.
            Continuing activity could result in heat stroke.
        3 = Danger: Heat cramps and heat exhaustion are likely.
            Heat stroke is probable with continued activity.
        4 = Extreme danger: Heat stroke is imminent.

    Args:
        hi: Heat index [C]

    Returns:
        category: An integer indicating the level of warning associated with the
            heat index. Values are one of the following:
                0 = No Warning
                1 = Caution
                2 = Extreme Caution
                3 = Danger
                4 = Extreme Danger
    """
    if hi < 26.6:
        category = 0
    elif hi < 32.2:
        category = 1
    elif hi < 40.5:
        category = 2
    elif hi < 54.4:
        category = 3
    else:
        category = 4

    return category
